from tqdm import tqdm

import numpy as np
from sentence_transformers import SentenceTransformer
import torch

from wasabi import msg, Printer

msg_print = Printer()


def chunk_list(orig_list, n_chunks):
    for i in range(0, len(orig_list), n_chunks):
        yield orig_list[i : i + n_chunks]


def load_bert(bert_model_name="sentence-transformers/all-MiniLM-L6-v2"):

    with msg_print.loading("Loading BERT model"):
        device = torch.device(f"cuda:0" if torch.cuda.is_available() else "cpu")
        bert_model = SentenceTransformer(bert_model_name, device=device)
        bert_model.max_seq_length = 512
    msg.good("BERT model loaded")
    return bert_model


def get_embeddings(
    text_list: list,
    bert_model,
    embed_chunk_size: int = 500,
    batch_size: int = 32,
) -> np.array:
    """
    Get embeddings for a list of texts

    Args:
        text_list (list): A lists of texts.
        bert_model: An initialised SentenceTransformer BERT model.
        embed_chunk_size (int): The number of texts per chunk to process.
        batch_size (int): BERT batch_size.
    Returns:
        np.array: The embeddings for the input list of texts
    """

    msg.info(
        f"Finding embeddings for {len(text_list)} texts chunked into {round(len(text_list)/embed_chunk_size)} chunks"
    )
    all_embeddings = []
    for batch_texts in tqdm(chunk_list(text_list, embed_chunk_size)):
        all_embeddings.append(
            bert_model.encode(np.array(batch_texts), batch_size=batch_size)
        )
    all_embeddings = np.concatenate(all_embeddings)
    msg.good("Texts embedded.")

    return all_embeddings



import pandas as pd

import re

from utils.utils import soc_mapper_config


def load_job_title_soc(soc_mapper_config: dict = soc_mapper_config) -> pd.DataFrame():
    """
    Load the ONS dataset which gives SOC codes for thousands of job titles
    """

    jobtitle_soc_data = pd.read_excel(
        soc_mapper_config["soc_data"]["soc_dir"],
        sheet_name=soc_mapper_config["soc_data"]["sheet_name"],
        converters={
            soc_mapper_config["soc_data"]["soc_2020_ext_col"]: str,
            soc_mapper_config["soc_data"]["soc_2020_col"]: str,
            soc_mapper_config["soc_data"]["soc_2010_col"]: str,
        },
    )

    return jobtitle_soc_data


def process_job_title_soc(
    jobtitle_soc_data: pd.DataFrame(), soc_mapper_config: dict = soc_mapper_config
) -> pd.DataFrame():
    """Standardise the column names for use in soc_map.py
    Args:
        jobtitle_soc_data (pd.DataFrame): the raw ONS SOC coding index dataset
    Returns:
        pd.DataFrame: the cleaned ONS SOC coding index dataset
    """

    jobtitle_soc_data = jobtitle_soc_data.rename(
        columns={
            soc_mapper_config["soc_data"]["soc_2020_ext_col"]: "SOC_2020_EXT",
            soc_mapper_config["soc_data"]["soc_2020_col"]: "SOC_2020",
            soc_mapper_config["soc_data"]["soc_2010_col"]: "SOC_2010",
            soc_mapper_config["soc_data"][
                "natural_order_col"
            ]: "INDEXOCC NATURAL WORD ORDER",
            soc_mapper_config["soc_data"]["sug_col"]: "SUB-UNIT GROUP DESCRIPTIONS",
            soc_mapper_config["soc_data"]["ug_col"]: "SOC 2020 UNIT GROUP DESCRIPTIONS",
            soc_mapper_config["soc_data"]["add_col"]: "ADD",
            soc_mapper_config["soc_data"]["ind_col"]: "IND",
        }
    )

    # Clean
    jobtitle_soc_data = jobtitle_soc_data[jobtitle_soc_data["SOC_2020"] != "}}}}"]

    return jobtitle_soc_data


def unique_soc_job_titles(jobtitle_soc_data: pd.DataFrame()) -> dict:
    """
    Taking the dataset of job titles and which SOC they belong to - create a unique
    dictionary where each key is a job title and the value is the SOC code.
    There are additional words to include in the job title if at first
    it is not unique.

    Args:
        jobtitle_soc_data (pd.DataFrame): the cleaned ONS SOC coding index dataset.

    Returns:
        dict: A dictionary where each key is a job title and the value is the SOC code.

    """

    col_name_0 = "INDEXOCC NATURAL WORD ORDER"
    col_name_1 = "ADD"
    col_name_2 = "IND"

    jobtitle_soc_data[f"{col_name_0} and {col_name_1}"] = jobtitle_soc_data.apply(
        lambda x: (
            x[col_name_0] + " " + x[col_name_1]
            if pd.notnull(x[col_name_1])
            else x[col_name_0]
        ),
        axis=1,
    )
    jobtitle_soc_data[f"{col_name_0} and {col_name_1} and {col_name_2}"] = (
        jobtitle_soc_data.apply(
            lambda x: (
                x[f"{col_name_0} and {col_name_1}"] + " " + x[col_name_2]
                if pd.notnull(x[col_name_2])
                else x[f"{col_name_0} and {col_name_1}"]
            ),
            axis=1,
        )
    )

    # Try to find a unique job title to SOC 2020 4 or 6 code mapping
    job_title_2_soc6_4 = {}
    for job_title, grouped_soc_data in jobtitle_soc_data.groupby(col_name_0):
        if grouped_soc_data["SOC_2020_EXT"].nunique() == 1:
            job_title_2_soc6_4[job_title] = (
                grouped_soc_data["SOC_2020_EXT"].unique()[0],
                grouped_soc_data["SOC_2020"].unique()[0],
                grouped_soc_data["SOC_2010"].unique()[0],
            )
        else:
            for job_title_1, grouped_soc_data_1 in grouped_soc_data.groupby(
                f"{col_name_0} and {col_name_1}"
            ):
                if grouped_soc_data_1["SOC_2020_EXT"].nunique() == 1:
                    job_title_2_soc6_4[job_title_1] = (
                        grouped_soc_data_1["SOC_2020_EXT"].unique()[0],
                        grouped_soc_data_1["SOC_2020"].unique()[0],
                        grouped_soc_data_1["SOC_2010"].unique()[0],
                    )
                else:
                    for (
                        job_title_2,
                        grouped_soc_data_2,
                    ) in grouped_soc_data_1.groupby(
                        f"{col_name_0} and {col_name_1} and {col_name_2}"
                    ):
                        if grouped_soc_data_2["SOC_2020_EXT"].nunique() == 1:
                            job_title_2_soc6_4[job_title_2] = (
                                grouped_soc_data_2["SOC_2020_EXT"].unique()[0],
                                grouped_soc_data_2["SOC_2020"].unique()[0],
                                grouped_soc_data_2["SOC_2010"].unique()[0],
                            )

    return job_title_2_soc6_4


def unique_soc_descriptions(soc_data: pd.DataFrame()) -> dict:
    """
    Taking the dataset of SOC and their descriptions - create a unique
    dictionary where each key is a description and the value is the SOC code.

    Args:
        soc_data (pd.DataFrame): the cleaned ONS SOC coding index dataset.

    Returns:
        dict: A dictionary where each key is a SOC description and the value is the SOC code.

    """
    soc_data["SUB-UNIT GROUP DESCRIPTIONS"] = soc_data[
        "SUB-UNIT GROUP DESCRIPTIONS"
    ].apply(lambda x: x.replace(" n.e.c.", "").replace(" n.e.c", ""))

    dd = soc_data[
        ["SUB-UNIT GROUP DESCRIPTIONS", "SOC_2020_EXT", "SOC_2020", "SOC_2010"]
    ].drop_duplicates()

    # There can be multiple 2010 codes for each 6 digit, so just output the most common
    soc_desc_2_code = {}
    for description, soc_info in dd.groupby("SUB-UNIT GROUP DESCRIPTIONS"):
        soc_2020_6 = soc_info["SOC_2020_EXT"].value_counts().index[0]
        soc_2020_4 = soc_info["SOC_2020"].value_counts().index[0]
        soc_2010 = list(soc_info["SOC_2010"].unique())
        soc_desc_2_code[description] = (soc_2020_6, soc_2020_4, soc_2010)

    return soc_desc_2_code


major_places = [
    "Central London",
    "Midlands",
    "London",
    "Birmingham",
    "Leeds",
    "Glasgow",
    "Sheffield",
    "Bradford",
    "Manchester",
    "Edinburgh",
    "Liverpool",
    "Bristol",
    "Cardiff",
    "Coventry",
    "Nottingham",
    "Leicester",
    "Sunderland",
    "Belfast",
    "Newcastle",
    "Brighton",
    "Hull",
    "Plymouth",
    "Carlisle",
    "Berkshire",
    "Doncaster",
    "Bedford",
    "Chichester",
    "Wakefield",
]
lower_case_end_words = [
    "nights",
    "part time",
    "full time",
    "hybrid",
    "maternity cover",
    "remote",
    "self employed",
    "work from home",
    "benefits",
    "flexible",
    "Office Based",
]

lower_case_all_end_words = [
    word.lower() for word in major_places + lower_case_end_words
]


def job_title_cleaner(
    text: str, lower_case_all_end_words: list = lower_case_all_end_words
) -> str:
    """
    Will apply a bunch of cleaning to a job title
    - removing certain things (locations or work type after a "-")
    - fixes some unicode &#163; -> £
    - Removes text after "£""

    Assumption: weird bad stuff comes after dashes or £ signs.
    So this won't work well for e.g "£30k Data Scientist" or "Remote - London Data Scientist"

    This isn't perfect, but should hopefully help quite a few examples

    Examples:
    'Part Home Based Block Manager - Chichester' -> 'Part Home Based Block Manager'
    'Employment Solicitor - Claimant - Leeds' -> 'Employment Solicitor - Claimant'
    'Retail Customer Service CSM 16hrs' -> 'Retail Customer Service CSM'
    'Bike Delivery Driver - London' -> 'Bike Delivery Driver'
    'Fulfillment Associate - &#163;1000 Sign on Bonus!' -> 'Fulfillment Associate'

    Args:
        text (str): the text of the job title you want to clean
        lower_case_all_end_words (list): a list of all the words to clean out
        if they are at the end of the job title.
    Returns:
        str: the cleaned job title

    """
    if text:
        text = str(text)

        findreplace = {
            "&amp;": " and ",
            "&#160;": " ",
            "&#163;": "£",
            "(part time)": " ",
        }
        for f, r in findreplace.items():
            text = text.replace(f, r)
        # Get rid of any double + spaces
        text = re.sub(r"\s{2,}", " ", text).strip()

        # Remove mentions of hours e.g. Customer Service 30hrs -> Customer Service
        text = re.sub(r"\d+\s*hrs", "", text).strip()

        # If there is a "£" remove everything after it (e.g. £30k per annum)
        # Unless it occurs very early in the text
        if "£" in text:
            matches = re.findall(r"£", text)
            index_found = text.index(matches[0])
            if index_found > 4:
                text = " ".join(text.split("£")[0:-1]).strip() if "£" in text else text

        # Remove certain things after the last dash
        if " - " in text:
            last_bit = text.split(" - ")[-1].strip().lower()
            # if any of the target words are in this, then remove everything after the dash
            # e.g "Data Scientist - remote, London" -> "Data Scientist"
            found = False
            for word in lower_case_all_end_words:
                if word in last_bit:
                    found = True
                    break
            if last_bit == "":  # This may happen if a £ was found
                found = True
            if found:
                # Remove everything after the lastedash
                text = " - ".join(text.split(" - ")[0:-1]).strip()

        if text:  # The cleaning may make it so we are left with nothing
            if text[-1] == "-":
                text = text[0:-1].strip()

    return text



"""
A class to map inputted job titles to their most likely SOC 2020 4-digit codes.

Usage:

from soc_mapper.soc_map import SOCMapper

soc_mapper = SOCMapper()
soc_mapper.load()
job_titles=["data scientist", "Assistant nurse", "Senior financial consultant - London"]

soc_mapper.get_soc(job_titles, return_soc_name=True)

"""

from collections import Counter
import os
from typing import List, Union

import pandas as pd
from tqdm import tqdm
import numpy as np

from nlp_link.soc_mapper.soc_map_utils import (
    load_job_title_soc,
    process_job_title_soc,
    job_title_cleaner,
    unique_soc_job_titles,
)
from nlp_link.linker import NLPLinker

from nlp_link.linker_utils import load_bert

from utils.utils import (
    soc_mapper_config,
    load_s3_json,
    load_local_json,
    save_to_s3,
    save_json_dict,
)

from wasabi import msg, Printer

msg_print = Printer()


class SOCMapper(object):
    """Class for linking job titles to SOC codes.

    The input job title is matched to a dataset of job titles with their 2020 SOC.
    - If the most similar job title is very similar, then the corresponding 6-digit SOC is outputted.
    - Otherwise, we look at a group of the most similar job titles, and if they all have the same 4-digit SOC, then this is outputted.

    ----------

    Args:
        soc_dir (str): The directory of the SOC coding index xlsx file.
        match_top_n (int): The number of most similar SOC matches to consider when calculating the final SOC and outputing
        sim_threshold (float): The similarity threshold for outputting the most similar SOC match.
        top_n_sim_threshold (float): The similarity threshold for a match being added to a group of SOC matches.
        minimum_n (int): The minimum size of a group of SOC matches.
        minimum_prop (float): If a group of SOC matches have a high proportion (>= minimum_prop) of the same SOC being matched, then use this SOC.

    ----------
    Methods
    ----------

    load_process_soc_data():
        Load the SOC data
    load(reset_embeddings=False, save_embeds=False):
            Load everything to use this class, recalculate SOC embeddings and save if desired
    find_most_similar_matches(job_titles, job_title_embeddings):
            Using the inputted job title embeddings and the SOC embeddings, find the full information about the most similar SOC job titles
    find_most_likely_soc(match_row):
            For the full match information for one job title, find the most likely SOC (via top match, or group of top matches)
    get_soc(job_titles, additional_info=False):
            (main function) For inputted job titles, output the best SOC match, add extra information about matches using the additional_info argument

    ----------
    Usage
    ----------
        from soc_mapper.soc_map import SOCMapper
        soc_mapper = SOCMapper()
        soc_mapper.load()
        matches = soc_mapper.get_soc(job_titles=["data scientist", "Assistant nurse", "Senior financial consultant - London"])
        >>> [(('2433/02', '2433', '2425'), 'data scientist'), (('6131/99', '6131', '6141'), 'assistant nurse'), (('2422/02', '2422', '3534'), 'financial consultant')]
    """

    def __init__(
        self,
        soc_dir: str = soc_mapper_config["soc_data"]["soc_dir"],
        match_top_n: int = soc_mapper_config["soc_mapper"]["match_top_n"],
        sim_threshold: float = soc_mapper_config["soc_mapper"]["sim_threshold"],
        top_n_sim_threshold: float = soc_mapper_config["soc_mapper"][
            "top_n_sim_threshold"
        ],
        minimum_n: int = soc_mapper_config["soc_mapper"]["minimum_n"],
        minimum_prop: float = soc_mapper_config["soc_mapper"]["minimum_prop"],
    ):
        self.soc_dir = soc_dir
        self.match_top_n = match_top_n
        self.sim_threshold = sim_threshold
        self.top_n_sim_threshold = top_n_sim_threshold
        self.minimum_n = minimum_n
        self.minimum_prop = minimum_prop

        self.soc_mapper_config = soc_mapper_config  # This is so a user could change the soc_data values easily if needed

    def load_process_soc_data(self):
        """
        Load the job titles to SOC codes dataset as found on the ONS website.
        A small amount of processing.
        """

        jobtitle_soc_data = process_job_title_soc(
            load_job_title_soc(soc_mapper_config=self.soc_mapper_config),
            soc_mapper_config=self.soc_mapper_config,
        )

        return jobtitle_soc_data

    def load(
        self,
        reset_embeddings: bool = soc_mapper_config["soc_mapper"]["reset_embeddings"],
        save_embeds: bool = False,
    ):
        """
        Load the BERT model, SOC coding index data, and load or calculate embeddings for the job titles in this dataset.
        Args:
            reset_embeddings (bool): Whether to re-calculate and save soc coding index embeddings or not. Will be done anyway if
                an embeddings file isn't found.
            save_embeds (bool): Whether to save the out the embeddings or not (only used if the embeddings weren't loaded)
        """

        self.nlp_link = NLPLinker()
        self.nlp_link.bert_model = load_bert()

        self.jobtitle_soc_data = self.load_process_soc_data()

        self.soc_2020_6_dict = dict(
            zip(
                self.jobtitle_soc_data["SOC_2020_EXT"],
                self.jobtitle_soc_data["SUB-UNIT GROUP DESCRIPTIONS"],
            )
        )
        self.soc_2020_4_dict = dict(
            zip(
                self.jobtitle_soc_data["SOC_2020"],
                self.jobtitle_soc_data["SOC 2020 UNIT GROUP DESCRIPTIONS"],
            )
        )
        self.job_title_2_soc6_4 = unique_soc_job_titles(self.jobtitle_soc_data)

        embeddings_output_dir = os.path.dirname(self.soc_dir)

        if "s3://" in embeddings_output_dir:
            s3_bucket_name = embeddings_output_dir.split("s3://")[1].split("/")[0]
            embeddings_output_s3_folder = "/".join(
                embeddings_output_dir.split("s3://")[1].split("/")[1:]
            )
            embeddings_path = os.path.join(
                embeddings_output_s3_folder, "soc_job_embeddings.json"
            )
            job_titles_path = os.path.join(
                embeddings_output_s3_folder, "soc_job_embeddings_titles.json"
            )
        else:
            s3_bucket_name = None
            embeddings_path = os.path.join(
                embeddings_output_dir, "soc_job_embeddings.json"
            )
            job_titles_path = os.path.join(
                embeddings_output_dir, "soc_job_embeddings_titles.json"
            )

        try:
            if not reset_embeddings:
                try:
                    if s3_bucket_name:
                        with msg_print.loading(
                            f"Loading SOC job title embeddings from S3 ..."
                        ):
                            self.all_soc_embeddings = load_s3_json(
                                s3_bucket_name, embeddings_path
                            )
                            self.soc_job_titles = load_s3_json(
                                s3_bucket_name, job_titles_path
                            )
                    else:
                        with msg_print.loading(
                            f"Loading SOC job title embeddings locally ..."
                        ):
                            self.all_soc_embeddings = load_local_json(embeddings_path)
                            self.soc_job_titles = load_local_json(job_titles_path)
                    msg.good("SOC job title embeddings loaded.")
                except:
                    msg.warn(f"SOC job title embeddings not found.")
                    raise
            else:
                raise
        except:
            msg.info(f"Calculating SOC job title embeddings ...")

            # Embed the SOC job titles
            self.soc_job_titles = list(self.job_title_2_soc6_4.keys())

            self.all_soc_embeddings = self.nlp_link._get_embeddings(self.soc_job_titles)

            if save_embeds:
                msg.info(f"Saving SOC job title embeddings")
                if s3_bucket_name:
                    save_to_s3(s3_bucket_name, self.all_soc_embeddings, embeddings_path)
                    save_to_s3(s3_bucket_name, self.soc_job_titles, job_titles_path)
                    msg.good(f"Saved to s3://{s3_bucket_name} + {embeddings_path} ...")
                else:
                    save_json_dict(self.all_soc_embeddings, embeddings_path)
                    save_json_dict(self.soc_job_titles, job_titles_path)
                    msg.good(f"Saved to {embeddings_path} ...")
            else:
                msg.warn(
                    f"Newly calculated SOC job title embeddings were not saved, set save_embeds=True if you'd like to save them to speed up future use."
                )

    def find_most_similar_matches(
        self,
        job_titles: Union[str, List[str]],
        job_title_embeddings: np.array(object),
    ) -> list:
        """
        Using the job title embeddings and the SOC job title embeddings,
        find the top n SOC job titles which are most similar to each input job title.

        Args:
            job_titles (str or list of strings): One or a list of inputted job titles.
            job_title_embeddings (np.array()): The embeddings for the inputted job titles.

        Outputs:
            list: A list of the most similar SOC data for each inputted job title.
        """

        matches_topn_dict = self.nlp_link.get_matches(
            input_data_ids=list(range(len(job_titles))),
            input_embeddings=job_title_embeddings,
            comparison_data_ids=list(range(len(self.all_soc_embeddings))),
            comparison_embeddings=self.all_soc_embeddings,
            top_n=self.match_top_n,
        )

        job_top_soc_matches = []
        for k, v in matches_topn_dict.items():
            top_soc_matches = []
            for top_match in v:
                soc_ix = top_match[0]
                similarity = top_match[1]
                soc_text = self.soc_job_titles[soc_ix]
                top_soc_matches.append(
                    [
                        soc_text,
                        self.job_title_2_soc6_4[soc_text][0],  # 6 digit
                        self.job_title_2_soc6_4[soc_text][1],  # 4 digit
                        self.job_title_2_soc6_4[soc_text][2],  # 2010 4 digit
                        similarity,
                    ]
                )
            job_top_soc_matches.append(
                {
                    "job_title": job_titles[k],
                    "top_soc_matches": top_soc_matches,
                }
            )

        return job_top_soc_matches

    def find_most_likely_soc(
        self,
        match_row: dict,
    ) -> tuple:
        """
        For a single job title and the details of the most similar SOC matches, find a single most likely SOC
        1. If the top match has a really high similarity score (>sim_threshold) at the 6-digit level then use this.
                This will return (soc, job_title)
        2. Get the 4-digit SOCs of the good (>top_n_sim_threshold) matches in the top n most similar.
        3. If there are a few of these (>=minimum_n) and over a certain proportion (>minimum_prop) of these are the same at the 4 digit level - use this as the SOC.
                This will return (soc, the job titles given for this same soc)

        Returns data in the format ((soc_2020_6, soc_2020_4, soc_2010), job_title) or None
        If pathway 1. (above) isn't true then the output will be ((None, soc_2020_4, None), job_title) and job_title will be a set of multiple

        Args:
            match_row: One element from the list outputted in find_most_similar_matches.
            e.g. {"job_title": 'principal data scientist', "top_soc_matches": [["data scientist", 6digit SOC, 4 digit SOC, 4 digit 2010 SOC, similarity_score], ...]}

        Output:
            tuple, None: Details of the most likely SOC match for this job title.
        """

        top_soc_match = match_row["top_soc_matches"][0][0]
        top_soc_match_code = (
            match_row["top_soc_matches"][0][1],
            match_row["top_soc_matches"][0][2],
            match_row["top_soc_matches"][0][3],
        )  # 6 digit, 4 digit, 4 2010
        top_soc_match_score = match_row["top_soc_matches"][0][4]  # The score

        if top_soc_match_score > self.sim_threshold:
            return (top_soc_match_code, top_soc_match)
        else:
            all_good_socs = [
                t[2]  # 4 digit 2020 SOC
                for t in match_row["top_soc_matches"]
                if t[4] > self.top_n_sim_threshold
            ]
            if len(all_good_socs) >= self.minimum_n:
                common_soc, num_common_soc = Counter(all_good_socs).most_common(1)[0]
                prop_most_common_soc = num_common_soc / len(all_good_socs)
                if prop_most_common_soc > self.minimum_prop:
                    return (
                        (None, common_soc, None),
                        set(
                            [
                                t[0]
                                for t in match_row["top_soc_matches"]
                                if (
                                    (t[4] > self.top_n_sim_threshold)
                                    and (t[2] == common_soc)
                                )
                            ]
                        ),
                    )
                else:
                    return None
            else:
                return None

    def get_soc(
        self,
        job_titles: Union[str, List[str]],
        additional_info: bool = False,
        return_soc_name: bool = False,
        clean_job_title: bool = True,
    ) -> list:
        """
        Get the most likely SOC for each inputted job title

        Args:
            job_titles (str, list): A single job title or a list of raw job titles
            additional_info (bool): Whether to provide additional information about the matches.
                        Return just the most likely soc match (False) or the top soc matches (True)
            return_soc_name (bool): Whether to output the SOC names of the most likely SOC (or just the codes).
                        When applied to lots of data this might not be as desirable.
            clean_job_title (bool): Whether to apply the cleaning function to the job title.

        Output:
            list: A list of the top matches for each job title inputted

        """

        if isinstance(job_titles, str):
            job_titles = [job_titles]

        # Clean the job titles
        if clean_job_title:
            job_titles = [job_title_cleaner(job_title) for job_title in job_titles]

        # Embed the input job titles
        job_title_embeddings = self.nlp_link._get_embeddings(job_titles)

        top_soc_matches = self.find_most_similar_matches(
            job_titles, job_title_embeddings
        )

        msg.info(f"Finding most likely SOC")
        found_count = 0
        for job_matches in top_soc_matches:
            most_likely_soc = self.find_most_likely_soc(job_matches)
            if most_likely_soc:
                ((soc_2020_6, soc_2020_4, soc_2010_4), job_title) = most_likely_soc
                if return_soc_name:
                    job_matches["most_likely_soc"] = (
                        (
                            (soc_2020_6, self.soc_2020_6_dict.get(soc_2020_6)),
                            (soc_2020_4, self.soc_2020_4_dict.get(soc_2020_4)),
                            soc_2010_4,
                        ),
                        job_title,
                    )
                else:
                    job_matches["most_likely_soc"] = (
                        (soc_2020_6, soc_2020_4, soc_2010_4),
                        job_title,
                    )
            else:
                job_matches["most_likely_soc"] = None
            if most_likely_soc:
                found_count += 1

        msg.good(
            f"Found SOCs for {found_count*100/len(top_soc_matches)}% of the job titles"
        )

        if additional_info:
            return top_soc_matches
        else:
            return [
                job_matches.get("most_likely_soc") for job_matches in top_soc_matches
            ]



"""
Class to link two datasets.

Example usage:

from nlp_link.linker import NLPLinker

nlp_link = NLPLinker()

# dict inputs
comparison_data = {'a': 'cats', 'b': 'dogs', 'd': 'rats', 'e': 'birds'}
input_data = {'x': 'owls', 'y': 'feline', 'z': 'doggies', 'za': 'dogs', 'zb': 'chair'}
nlp_link.load(comparison_data)
matches = nlp_link.link_dataset(input_data)
# Top match output
print(matches)

# list inputs
comparison_data = ['cats', 'dogs', 'rats', 'birds']
input_data = ['owls', 'feline', 'doggies', 'dogs','chair']
nlp_link.load(comparison_data)
matches = nlp_link.link_dataset(input_data)
# Top match output
print(matches)

"""

from tqdm import tqdm
from sklearn.metrics.pairwise import cosine_similarity
import numpy as np
import pandas as pd

from typing import Union, Optional

from nlp_link.linker_utils import chunk_list, get_embeddings, load_bert

from wasabi import msg, Printer

msg_print = Printer()

# TO DO: cosine or euclidean?


class NLPLinker(object):
    """docstring for NLPLinker"""

    def __init__(self, batch_size=32, embed_chunk_size=500, match_chunk_size=10000):
        super(NLPLinker, self).__init__()
        self.batch_size = batch_size
        self.embed_chunk_size = embed_chunk_size
        self.match_chunk_size = match_chunk_size
        ## Cleaning?

    def _process_dataset(
        self,
        input_data: Union[list, dict, pd.DataFrame],
        id_column: Optional[str] = None,
        text_column: Optional[str] = None,
    ) -> dict:
        """Check and process a dataset according to the input type
        Args:
            input_data (Union[list, dict, pd.DataFrame])
                A list of texts or a dictionary of texts where the key is the unique id.
                If a list is given then a unique id will be assigned with the index order.

        Returns:
            dict: key is the id and the value is the text
        """

        if isinstance(input_data, list):
            return {ix: text for ix, text in enumerate(input_data)}
        elif isinstance(input_data, dict):
            return input_data
        elif isinstance(input_data, pd.DataFrame):
            try:
                return dict(zip(input_data[id_column], input_data[text_column]))
            except:
                msg.warn(
                    "Input is a dataframe, please specify id_column and text_column"
                )
        else:
            msg.warn(
                "The input_data input must be a dictionary, a list or pandas dataframe"
            )

        if not isinstance(input_data[0], str):
            msg.warn(
                "The input_data input must be a list of texts, or a dictionary where the values are texts"
            )

    def load(
        self,
        comparison_data: Union[list, dict],
    ):
        """
        Load the embedding model and embed the comparison dataset
        Args:
            comparison_data (Union[list, dict]): The comparison texts to find links to.
                A list of texts or a dictionary of texts where the key is the unique id.
                If a list is given then a unique id will be assigned with the index order.
        """
        self.bert_model = load_bert()

        self.comparison_data = self._process_dataset(comparison_data)
        self.comparison_data_texts = list(self.comparison_data.values())
        self.comparison_data_ids = list(self.comparison_data.keys())

        self.comparison_embeddings = self._get_embeddings(self.comparison_data_texts)

    def _get_embeddings(self, text_list: list) -> np.array:
        """
        Get embeddings for a list of texts

        Args:
            text_list (list): A lists of texts
        Returns:
            np.array: The embeddings for the input list of texts
        """

        return get_embeddings(
            text_list=text_list,
            embed_chunk_size=self.embed_chunk_size,
            batch_size=self.batch_size,
            bert_model=self.bert_model,
        )

    def get_matches(
        self,
        input_data_ids: list,
        input_embeddings: np.array,
        comparison_data_ids: list,
        comparison_embeddings: np.array,
        top_n: int,
        drop_most_similar: bool = False,
    ) -> dict:
        """
        Find top matches across two datasets using their embeddings.

        Args:
            input_data_ids (list): The ids of the input texts.
            input_embeddings (np.array): Embeddings for the input texts.
            comparison_data_ids (list): The ids of the comparison texts.
            comparison_embeddings (np.array): Embeddings for the comparison texts.
            top_n (int): The number of top links to return in the output.
            drop_most_similar (bool, default = False): Whether to not output the most similar match, this would be set to True if you are matching a list with itself.

        Returns:
            dict: The top matches for each input id.
        """

        msg.info(
            f"Finding the top dataset matches for {len(input_data_ids)} input texts chunked into {round(len(input_data_ids)/self.match_chunk_size)}"
        )

        if drop_most_similar:
            top_n = top_n + 1
            start_n = 1
        else:
            start_n = 0

        # We chunk up comparisons otherwise it can crash
        matches_topn = {}
        for batch_indices in tqdm(
            chunk_list(range(len(input_data_ids)), n_chunks=self.match_chunk_size)
        ):
            batch_input_ids = [input_data_ids[i] for i in batch_indices]
            batch_input_embeddings = [input_embeddings[i] for i in batch_indices]

            batch_similarities = cosine_similarity(
                batch_input_embeddings, comparison_embeddings
            )

            # Top links for each input text
            for input_ix, similarities in enumerate(batch_similarities):
                top_links = []
                for comparison_ix in np.flip(np.argsort(similarities))[start_n:top_n]:
                    # comparison data id + cosine similarity score
                    top_links.append(
                        [
                            comparison_data_ids[comparison_ix],
                            similarities[comparison_ix],
                        ]
                    )
                matches_topn[batch_input_ids[input_ix]] = top_links
        return matches_topn

    def link_dataset(
        self,
        input_data: Union[list, dict],
        top_n: int = 3,
        format_output: bool = True,
        drop_most_similar: bool = False,
    ) -> dict:
        """
        Link a dataset to the comparison dataset.

        Args:
            input_data (Union[list, dict]): The main dictionary to be linked to texts in the loaded comparison_data.
                A list of texts or a dictionary of texts where the key is the unique id.
                If a list is given then a unique id will be assigned with the index order.
            top_n (int, default = 3): The number of top links to return in the output.
            format_output (bool, default = True): If you'd like the output to be formatted to include the texts of
                the matched datasets or not (will just give the indices).
            drop_most_similar (bool, default = False): Whether to not output the most similar match, this would be set to True if you are matching a list with itself.
        Returns:
            dict: The keys are the ids of the input_data and the values are a list of lists of the top_n most similar
                ids from the comparison_data and a probability score.
                e.g. {'x': [['a', 0.75], ['c', 0.7]], 'y': [...]}
        """

        try:
            msg.info(
                f"Comparing {len(input_data)} input texts to {len(self.comparison_embeddings)} comparison texts"
            )
        except:
            msg.warning(
                "self.comparison_embeddings does not exist - you may have not run load()"
            )

        input_data = self._process_dataset(input_data)
        input_data_texts = list(input_data.values())
        input_data_ids = list(input_data.keys())

        input_embeddings = self._get_embeddings(input_data_texts)

        self.matches_topn = self.get_matches(
            input_data_ids,
            input_embeddings,
            self.comparison_data_ids,
            self.comparison_embeddings,
            top_n,
            drop_most_similar,
        )

        if format_output:
            # Format the output into a user friendly pandas format with the top link only
            df_output = pd.DataFrame(
                [
                    {
                        "input_id": input_id,
                        "input_text": input_data[input_id],
                        "link_id": link_data[0][0],
                        "link_text": self.comparison_data[link_data[0][0]],
                        "similarity": link_data[0][1],
                    }
                    for input_id, link_data in self.matches_topn.items()
                ]
            )
            return df_output
        else:
            return self.matches_topn
